/* ----------------------------- MNI Header -----------------------------------
@NAME       : flipdef
@INPUT      : argc, argv - command line arguments
@OUTPUT     : (none)
@RETURNS    : status
@DESCRIPTION: Program designed to invert the representation of 
        a deformation field to speed up resampling into the target
	space. 
@METHOD     : 
@GLOBALS    : 
@CALLS      : 
@CREATED    : Mon May 29 09:07:14 MET DST 1995 Collins
@MODIFIED   : $Log:	 $

@COPYRIGHT  :
              Copyright 1995 Louis Collins, McConnell Brain Imaging Centre, 
              Montreal Neurological Institute, McGill University.
              Permission to use, copy, modify, and distribute this
              software and its documentation for any purpose and without
              fee is hereby granted, provided that the above copyright
              notice appear in all copies.  The author and McGill University
              make no representations about the suitability of this
              software for any purpose.  It is provided "as is" without
              express or implied warranty.
---------------------------------------------------------------------------- */

#ifndef lint
static char rcsid[]="$Header: /usr/people/neelin/src/file/minc/progs/xfm/RCS/flipdef.c,v 1.4 93/10/12 12:52:37 neelin Exp Locker: neelin $";
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <volume_io.h>
#include <minc_def.h>

/* Constants */
#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif
#ifndef public
#  define public
#  define private static
#endif


void get_volume_XYZV_indices(Volume data, int xyzv[])
{
  
  int 
    axis, i, vol_dims;
  char 
    **data_dim_names;
 
  vol_dims       = get_volume_n_dimensions(data);
  data_dim_names = get_volume_dimension_names(data);
 
  for_less(i,0,N_DIMENSIONS+1) xyzv[i] = -1;
  for_less(i,0,vol_dims) {
    if (convert_dim_name_to_spatial_axis(data_dim_names[i], &axis )) {
      xyzv[axis] = i; 
    } 
    else {     /* not a spatial axis */
      xyzv[Z+1] = i;
    }
  }
  delete_dimension_names(data_dim_names);
 
}

public  void  collins_general_transform_point(
    General_transform   *transform,
    Real                x,
    Real                y,
    Real                z,
    Real                *x_transformed,
    Real                *y_transformed,
    Real                *z_transformed );

public  void  collins_general_inverse_transform_point(
    General_transform   *transform,
    Real                x,
    Real                y,
    Real                z,
    Real                *x_transformed,
    Real                *y_transformed,
    Real                *z_transformed );



/* Main program */
char *prog_name;

int main(int argc, char *argv[])
{
   General_transform 
     transform, 
     inverse, 
     *grid_transform_ptr, 
     forward_transform;
   Volume 
     volume;
   Real
     voxel[MAX_DIMENSIONS],
     wx,wy,wz, inv_x, inv_y, inv_z,
     def_values[MAX_DIMENSIONS];
   int 
     prog_count,
     sizes[MAX_DIMENSIONS],
     xyzv[MAX_DIMENSIONS],
     index[MAX_DIMENSIONS],
     i,j,k,
     n_trans,
     trans_count;
   progress_struct
     progress;

   prog_name = argv[0];

   /* Check arguments */
   if (argc != 3) {
      (void) fprintf(stderr, "Usage: %s <input.xfm> <result.xfm>\n",
                     argv[0]);
      exit(EXIT_FAILURE);
   }

   /* Read in file that has a def field to invert */
   if (input_transform_file(argv[1], &transform) != OK) {
      (void) fprintf(stderr, "%s: Error reading transform file %s\n",
                     argv[0], argv[1]);
      exit(EXIT_FAILURE);
   }

   for_less(trans_count,0, get_n_concated_transforms(&transform) ) {

     grid_transform_ptr = get_nth_general_transform(&transform, trans_count );
     
     if (grid_transform_ptr->type == GRID_TRANSFORM) {

       copy_general_transform(grid_transform_ptr,
			      &forward_transform);

       forward_transform.inverse_flag = !(forward_transform.inverse_flag);

       volume = grid_transform_ptr->displacement_volume;

       get_volume_sizes(volume, sizes);
       get_volume_XYZV_indices(volume,xyzv);

       for_less(i,0,MAX_DIMENSIONS) index[i] = 0;

       initialize_progress_report(&progress, FALSE, 
				  sizes[xyzv[X]]*sizes[xyzv[Y]]*sizes[xyzv[Z]]+1,
				  "Inverting def field");
       prog_count = 0;

       for_less( index[ xyzv[X] ], 0, sizes[ xyzv[X] ])
	 for_less( index[ xyzv[Y] ], 0, sizes[ xyzv[Y] ])
	   for_less( index[ xyzv[Z] ], 0, sizes[ xyzv[Z] ]) {
	     
	     index[ xyzv[Z+1] ] = 0;
	     for_less(i,0,MAX_DIMENSIONS) voxel[i] = (Real)index[i];
       
	     convert_voxel_to_world(volume, voxel, &wx, &wy, &wz);
	     
	     if (sizes[ xyzv[Z] ] ==1)
	       collins_grid_inverse_transform_point(&forward_transform,
					    wx, wy, wz,
					    &inv_x, &inv_y, &inv_z);
	     else
	       grid_inverse_transform_point(&forward_transform,
					    wx, wy, wz,
					    &inv_x, &inv_y, &inv_z);
	     def_values[X] = inv_x - wx;
	     def_values[Y] = inv_y - wy;
	     def_values[Z] = inv_z - wz;

	     for_less(index[ xyzv[Z+1] ],0,3)
	       set_volume_real_value(volume,
				     index[0],index[1],index[2],index[3],index[4],
				     def_values[ index[ xyzv[Z+1] ]]);

	     prog_count++;
	     update_progress_report(&progress, prog_count);
	   }
       
       terminate_progress_report(&progress);

       delete_general_transform(&forward_transform);

       grid_transform_ptr->inverse_flag = !(grid_transform_ptr->inverse_flag);
       
     }

   }
   

   /* Write out the transform */
   if (output_transform_file(argv[2], NULL, &transform) != OK) {
      (void) fprintf(stderr, "%s: Error writing transform file %s\n",
                     argv[0], argv[2]);
      exit(EXIT_FAILURE);
   }

   exit(EXIT_SUCCESS);
}
