# ------------------------------ MNI Header ----------------------------------
#@NAME       : misc_utilities.pl
#@DESCRIPTION: Miscellaneous and unclassifiable (but otherwise useful!)
#              utility routines:
#                 &timestamp
#                 &userstamp
#@GLOBALS    : 
#@CREATED    : 95/08/07, Greg Ward (from gpw_utilities.pl)
#@MODIFIED   : 
#@VERSION    : $Id: misc_utilities.pl,v 1.3 1996/02/12 19:08:28 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

# ------------------------------ MNI Header ----------------------------------
#@NAME       : timestamp
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Generates and returns a timestamp of the form 
#              [1995-05-16 22:30].
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/5/16, GW (from &doit)
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub timestamp
{
   local ($sec,$min,$hour,$mday,$mon,$year) = localtime (time);
   sprintf ("[%d-%02d-%02d %02d:%02d]", 
	    $year, $mon+1, $mday, $hour, $min);
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : userstamp
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Generates and returns a "userstamp" of the form [greg@panter].
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/5/16, GW
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub userstamp
{
   local ($user) = getpwuid ($<);
   local ($host) = `hostname`;
   chop $host;
   sprintf ("[%s@%s]", $user, $host);
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : comp_num_lists
#@INPUT      : $a1 - reference to first list
#              $a2 - reference to second list
#@OUTPUT     : 
#@RETURNS    : 0 if any mismatch between the two lists (i.e. they are of
#                different length, or any corresponding elements aren't
#                equal in the numeric sense)
#              1 if the two lists are numerically identical
#@DESCRIPTION: 
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 96/02/12 GPW (direct copyof comp_arrays in autocrop)
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub comp_num_lists
{
   die "comp_num_lists: wrong number of arguments" unless (@_ == 2);
   local ($a1, $a2) = @_;
   
   return 0 unless (@$a1 == @$a2);
   for $i (0 .. $#$a1)
   {
      return 0 unless ($a1->[$i] == $a2->[$i]);
   }
   return 1;
}

1;
